# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import base64
import io
import os
import re
import uuid

from .. import exception
from ..common import ContentLocation
from ..locale import gettext_lazy as _
from .base import AttachBase


class AttachMemory(AttachBase):
    """A wrapper for Memory based attachment sources."""

    # The default descriptive name associated with the service
    service_name = _("Memory")

    # The default protocol
    protocol = "memory"

    # Content is local to the same location as the apprise instance
    # being called (server-side)
    location = ContentLocation.LOCAL

    def __init__(
        self,
        content=None,
        name=None,
        mimetype=None,
        encoding="utf-8",
        **kwargs,
    ):
        """Initialize Memory Based Attachment Object."""
        # Create our BytesIO object
        self._data = io.BytesIO()

        if content is None:
            # Empty; do nothing
            pass

        elif isinstance(content, str):
            content = content.encode(encoding)
            if mimetype is None:
                mimetype = "text/plain"

            if not name:
                # Generate a unique filename
                name = str(uuid.uuid4()) + ".txt"

        elif not isinstance(content, bytes):
            raise TypeError(
                "Provided content for memory attachment is invalid"
            )

        # Store our content
        if content:
            self._data.write(content)

        if mimetype is None:
            # Default mimetype
            mimetype = "application/octet-stream"

        if not name:
            # Generate a unique filename
            name = str(uuid.uuid4()) + ".dat"

        # Initialize our base object
        super().__init__(name=name, mimetype=mimetype, **kwargs)

        return

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "mime": self._mimetype,
        }

        return "memory://{name}?{params}".format(
            name=self.quote(self._name),
            params=self.urlencode(params, safe="/"),
        )

    def open(self, *args, **kwargs):
        """Return our memory object."""
        # Return our object
        self._data.seek(0, 0)
        return self._data

    def __enter__(self):
        """Support with clause."""
        # Return our object
        self._data.seek(0, 0)
        return self._data

    def download(self, **kwargs):
        """Handle memory download() call."""

        if self.location == ContentLocation.INACCESSIBLE:
            # our content is inaccessible
            return False

        if self.max_file_size > 0 and len(self) > self.max_file_size:
            # The content to attach is to large
            self.logger.error(
                "Content exceeds allowable maximum memory size"
                f" ({int(self.max_file_size / 1024)}KB):"
                f" {self.url(privacy=True)}"
            )

            # Return False (signifying a failure)
            return False

        return True

    def base64(self, encoding="ascii"):
        """We need to over-ride this since the base64 sub-library seems to
        close our file descriptor making it no longer referencable."""

        if not self:
            # We could not access the attachment
            self.logger.error(
                f"Could not access attachment {self.url(privacy=True)}."
            )
            raise exception.AppriseFileNotFound("Attachment Missing")
        self._data.seek(0, 0)

        return (
            base64.b64encode(self._data.read()).decode(encoding)
            if encoding
            else base64.b64encode(self._data.read())
        )

    def invalidate(self):
        """Removes data."""
        self._data.truncate(0)
        return

    def exists(self):
        """Over-ride exists() call."""
        size = len(self)
        return bool(
            self.location != ContentLocation.INACCESSIBLE
            and size > 0
            and (
                self.max_file_size <= 0
                or (self.max_file_size > 0 and size <= self.max_file_size)
            )
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL so that we can handle all different file paths and
        return it as our path object."""

        results = AttachBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early; it's not a good URL
            return results

        if "name" not in results:
            # Allow fall-back to be from URL
            match = re.match(r"memory://(?P<path>[^?]+)(\?.*)?", url, re.I)
            if match:
                # Store our filename only (ignore any defined paths)
                results["name"] = os.path.basename(
                    AttachMemory.unquote(match.group("path"))
                )
        return results

    @property
    def path(self):
        """Return the filename."""
        if not self.exists():
            # we could not obtain our path
            return None

        return self._name

    def __len__(self):
        """Returns the size of he memory attachment."""
        return self._data.getbuffer().nbytes

    def __bool__(self):
        """Allows the Apprise object to be wrapped in an based 'if statement'.

        True is returned if our content was downloaded correctly.
        """

        return self.exists()



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from enum import Enum


class NotifyType(str, Enum):
    """A simple mapping of notification types most commonly used with all types
    of logging and notification services."""

    INFO = "info"
    SUCCESS = "success"
    WARNING = "warning"
    FAILURE = "failure"


# Define our types so we can verify if we need to
NOTIFY_TYPES: frozenset[str] = frozenset(e.value for e in NotifyType)


class NotifyImageSize(str, Enum):
    """A list of pre-defined image sizes to make it easier to work with defined
    plugins."""

    XY_32 = "32x32"
    XY_72 = "72x72"
    XY_128 = "128x128"
    XY_256 = "256x256"


# Define our image sizes so we can verify if we need to
NOTIFY_IMAGE_SIZES: frozenset[str] = \
    frozenset(e.value for e in NotifyImageSize)


class NotifyFormat(str, Enum):
    """A list of pre-defined text message formats that can be passed via the
    apprise library."""

    TEXT = "text"
    HTML = "html"
    MARKDOWN = "markdown"


# Define our formats so we can verify if we need to
NOTIFY_FORMATS: frozenset[str] = frozenset(e.value for e in NotifyFormat)


class OverflowMode(str, Enum):
    """A list of pre-defined modes of how to handle the text when it exceeds
    the defined maximum message size."""

    # Send the data as is; untouched.  Let the upstream server decide how the
    # content is handled.  Some upstream services might gracefully handle this
    # with expected intentions; others might not.
    UPSTREAM = "upstream"

    # Always truncate the text when it exceeds the maximum message size and
    # send it anyway
    TRUNCATE = "truncate"

    # Split the message into multiple smaller messages that fit within the
    # limits of what is expected.  The smaller messages are sent
    SPLIT = "split"


# Define our modes so we can verify if we need to
OVERFLOW_MODES: frozenset[str] = frozenset(e.value for e in OverflowMode)


class ConfigFormat(str, Enum):
    """A list of pre-defined config formats that can be passed via the apprise
    library."""

    # A text based configuration. This consists of a list of URLs delimited by
    # a new line.  pound/hashtag (#) or semi-colon (;) can be used as comment
    # characters.
    TEXT = "text"

    # YAML files allow a more rich of an experience when settig up your
    # apprise configuration files.
    YAML = "yaml"


# Define our configuration formats mostly used for verification
CONFIG_FORMATS: frozenset[str] = frozenset(e.value for e in ConfigFormat)


class ContentIncludeMode(str, Enum):
    """The different Content inclusion modes.

    All content based plugins will have one of these associated with it.
    """

    # - Content inclusion of same type only; hence a file:// can include
    #   a file://
    # - Cross file inclusion is not allowed unless insecure_includes (a flag)
    #   is set to True. In these cases STRICT acts as type ALWAYS
    STRICT = "strict"

    # This content type can never be included
    NEVER = "never"

    # This content can always be included
    ALWAYS = "always"


# Define our file inclusion types so we can verify if we need to
CONTENT_INCLUDE_MODES: frozenset[str] = \
    frozenset(e.value for e in ContentIncludeMode)


class ContentLocation(str, Enum):
    """This is primarily used for handling file attachments.  The idea is to
    track the source of the attachment itself.  We don't want remote calls to a
    server to access local attachments for example.

    By knowing the attachment type and cross-associating it with how we plan on
    accessing the content, we can make a judgement call (for security reasons)
    if we will allow it.

    Obviously local uses of apprise can access both local and remote type
    files.
    """

    # Content is located locally (on the same server as apprise)
    LOCAL = "local"

    # Content is located in a remote location
    HOSTED = "hosted"

    # Content is inaccessible
    INACCESSIBLE = "n/a"


# Define our location types so we can verify if we need to
CONTENT_LOCATIONS: frozenset[str] = frozenset(e.value for e in ContentLocation)


class PersistentStoreMode(str, Enum):
    # Allow persistent storage; write on demand
    AUTO = "auto"

    # Always flush every change to disk after it's saved. This has higher i/o
    # but enforces disk reflects what was set immediately
    FLUSH = "flush"

    # memory based store only
    MEMORY = "memory"


# Define our persistent storage modes so we can verify if we need to
PERSISTENT_STORE_MODES: frozenset[str] = \
    frozenset(e.value for e in PersistentStoreMode)


class PersistentStoreState(str, Enum):
    """Defines the persistent states describing what has been cached."""

    # Persistent Directory is actively cross-referenced against a matching URL
    ACTIVE = "active"

    # Persistent Directory is no longer being used or has no cross-reference
    STALE = "stale"

    # Persistent Directory is not utilizing any disk space at all, however
    # it potentially could if the plugin it successfully cross-references
    # is utilized
    UNUSED = "unused"


# Define our persistent storage states so we can verify if we need to
PERSISTENT_STORE_STATES: frozenset[str] = \
    frozenset(e.value for e in PersistentStoreState)

# This is a reserved tag that is automatically assigned to every
# Notification Plugin
MATCH_ALL_TAG = "all"

# Will cause notification to trigger under any circumstance even if an
# exclusive tagging was provided.
MATCH_ALWAYS_TAG = "always"



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Sign-up with https://dashboard.nexmo.com/
#
# Get your (api) key and secret here:
#   - https://dashboard.nexmo.com/getting-started-guide
#
import contextlib

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import is_phone_no, parse_phone_no, validate_regex
from .base import NotifyBase


class NotifyVonage(NotifyBase):
    """A wrapper for Vonage Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Vonage"

    # The services URL
    service_url = "https://dashboard.nexmo.com/"

    # The default protocol (nexmo kept for backwards compatibility)
    secure_protocol = ("vonage", "nexmo")

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_nexmo"

    # Vonage uses the http protocol with JSON requests
    notify_url = "https://rest.nexmo.com/sms/json"

    # The maximum length of the body
    body_maxlen = 160

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = (
        "{schema}://{apikey}:{secret}@{from_phone}",
        "{schema}://{apikey}:{secret}@{from_phone}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "required": True,
                "regex": (r"^[a-z0-9]+$", "i"),
                "private": True,
            },
            "secret": {
                "name": _("API Secret"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9]+$", "i"),
            },
            "from_phone": {
                "name": _("From Phone No"),
                "type": "string",
                "required": True,
                "regex": (r"^\+?[0-9\s)(+-]+$", "i"),
                "map_to": "source",
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "from_phone",
            },
            "key": {
                "alias_of": "apikey",
            },
            "secret": {
                "alias_of": "secret",
            },
            # Default Time To Live
            # By default Vonage attempt delivery for 72 hours, however the
            # maximum effective value depends on the operator and is typically
            # 24 - 48 hours. We recommend this value should be kept at its
            # default or at least 30 minutes.
            "ttl": {
                "name": _("ttl"),
                "type": "int",
                "default": 900000,
                "min": 20000,
                "max": 604800000,
            },
        },
    )

    def __init__(
        self, apikey, secret, source, targets=None, ttl=None, **kwargs
    ):
        """Initialize Vonage Object."""
        super().__init__(**kwargs)

        # API Key (associated with project)
        self.apikey = validate_regex(
            apikey, *self.template_tokens["apikey"]["regex"]
        )
        if not self.apikey:
            msg = f"An invalid Vonage API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # API Secret (associated with project)
        self.secret = validate_regex(
            secret, *self.template_tokens["secret"]["regex"]
        )
        if not self.secret:
            msg = f"An invalid Vonage API Secret ({secret}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Set our Time to Live Flag
        self.ttl = self.template_args["ttl"]["default"]
        with contextlib.suppress(ValueError, TypeError):
            # update our ttl if we're dealing with an integer
            self.ttl = int(ttl)

        if (
            self.ttl < self.template_args["ttl"]["min"]
            or self.ttl > self.template_args["ttl"]["max"]
        ):
            msg = f"The Vonage TTL specified ({self.ttl}) is out of range."
            self.logger.warning(msg)
            raise TypeError(msg)

        # The Source Phone #
        self.source = source

        result = is_phone_no(source)
        if not result:
            msg = (
                f"The Account (From) Phone # specified ({source}) is invalid."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our parsed value
        self.source = result["full"]

        # Parse our targets
        self.targets = []

        for target in parse_phone_no(targets):
            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
                continue

            # store valid phone number
            self.targets.append(result["full"])

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Vonage Notification."""

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded",
        }

        # Prepare our payload
        payload = {
            "api_key": self.apikey,
            "api_secret": self.secret,
            "ttl": self.ttl,
            "from": self.source,
            "text": body,
            # The to gets populated in the loop below
            "to": None,
        }

        # Create a copy of the targets list
        targets = list(self.targets)

        if len(targets) == 0:
            # No sources specified, use our own phone no
            targets.append(self.source)

        while len(targets):
            # Get our target to notify
            target = targets.pop(0)

            # Prepare our user
            payload["to"] = target

            # Some Debug Logging
            self.logger.debug(
                "Vonage POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"Vonage Payload: {payload}")

            # Always call throttle before any remote server i/o is made
            self.throttle()

            try:
                r = requests.post(
                    self.notify_url,
                    data=payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyVonage.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send Vonage notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(f"Sent Vonage notification to {target}.")

            except requests.RequestException as e:
                self.logger.warning(
                    f"A Connection error occurred sending Vonage:{target} "
                    "notification."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol[0], self.apikey, self.secret)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "ttl": str(self.ttl),
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{key}:{secret}@{source}/{targets}/?{params}".format(
            schema=self.secure_protocol[0],
            key=self.pprint(self.apikey, privacy, safe=""),
            secret=self.pprint(
                self.secret, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            source=NotifyVonage.quote(self.source, safe=""),
            targets="/".join(
                [NotifyVonage.quote(x, safe="") for x in self.targets]
            ),
            params=NotifyVonage.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifyVonage.split_path(results["fullpath"])

        # The hostname is our source number
        results["source"] = NotifyVonage.unquote(results["host"])

        # Get our account_side and auth_token from the user/pass config
        results["apikey"] = NotifyVonage.unquote(results["user"])
        results["secret"] = NotifyVonage.unquote(results["password"])

        # API Key
        if "key" in results["qsd"] and len(results["qsd"]["key"]):
            # Extract the API Key from an argument
            results["apikey"] = NotifyVonage.unquote(results["qsd"]["key"])

        # API Secret
        if "secret" in results["qsd"] and len(results["qsd"]["secret"]):
            # Extract the API Secret from an argument
            results["secret"] = NotifyVonage.unquote(results["qsd"]["secret"])

        # Support the 'from'  and 'source' variable so that we can support
        # targets this way too.
        # The 'from' makes it easier to use yaml configuration
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifyVonage.unquote(results["qsd"]["from"])
        if "source" in results["qsd"] and len(results["qsd"]["source"]):
            results["source"] = NotifyVonage.unquote(results["qsd"]["source"])

        # Support the 'ttl' variable
        if "ttl" in results["qsd"] and len(results["qsd"]["ttl"]):
            results["ttl"] = NotifyVonage.unquote(results["qsd"]["ttl"])

        # Support the 'to' variable so that we can support rooms this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyVonage.parse_phone_no(
                results["qsd"]["to"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Signup @ https://www.opsgenie.com
#
# Generate your Integration API Key
#   https://app.opsgenie.com/settings/integration/add/API/

# Knowing this, you can build your Opsgenie URL as follows:
#  opsgenie://{apikey}/
#  opsgenie://{apikey}/@{user}
#  opsgenie://{apikey}/*{schedule}
#  opsgenie://{apikey}/^{escalation}
#  opsgenie://{apikey}/#{team}
#
# You can mix and match what you want to notify freely
#  opsgenie://{apikey}/@{user}/#{team}/*{schedule}/^{escalation}
#
# If no target prefix is specified, then it is assumed to be a user.
#
# API Documentation: https://docs.opsgenie.com/docs/alert-api
# API Integration Docs: https://docs.opsgenie.com/docs/api-integration

import hashlib
from json import dumps, loads

import requests

from ..common import NOTIFY_TYPES, NotifyType, PersistentStoreMode
from ..locale import gettext_lazy as _
from ..utils.parse import is_uuid, parse_bool, parse_list, validate_regex
from .base import NotifyBase


class OpsgenieCategory(NotifyBase):
    """We define the different category types that we can notify."""

    USER = "user"
    SCHEDULE = "schedule"
    ESCALATION = "escalation"
    TEAM = "team"


OPSGENIE_CATEGORIES = (
    OpsgenieCategory.USER,
    OpsgenieCategory.SCHEDULE,
    OpsgenieCategory.ESCALATION,
    OpsgenieCategory.TEAM,
)


class OpsgenieAlertAction:
    """Defines the supported actions."""

    # Use mapping (specify :key=arg to over-ride)
    MAP = "map"

    # Create new alert (default)
    NEW = "new"

    # Close Alert
    CLOSE = "close"

    # Delete Alert
    DELETE = "delete"

    # Acknowledge Alert
    ACKNOWLEDGE = "acknowledge"

    # Add note to alert
    NOTE = "note"


OPSGENIE_ACTIONS = (
    OpsgenieAlertAction.MAP,
    OpsgenieAlertAction.NEW,
    OpsgenieAlertAction.CLOSE,
    OpsgenieAlertAction.DELETE,
    OpsgenieAlertAction.ACKNOWLEDGE,
    OpsgenieAlertAction.NOTE,
)

# Map all support Apprise Categories to Opsgenie Categories
OPSGENIE_ALERT_MAP = {
    NotifyType.INFO: OpsgenieAlertAction.CLOSE,
    NotifyType.SUCCESS: OpsgenieAlertAction.CLOSE,
    NotifyType.WARNING: OpsgenieAlertAction.NEW,
    NotifyType.FAILURE: OpsgenieAlertAction.NEW,
}


# Regions
class OpsgenieRegion:
    US = "us"
    EU = "eu"


# Opsgenie APIs
OPSGENIE_API_LOOKUP = {
    OpsgenieRegion.US: "https://api.opsgenie.com/v2/alerts",
    OpsgenieRegion.EU: "https://api.eu.opsgenie.com/v2/alerts",
}

# A List of our regions we can use for verification
OPSGENIE_REGIONS = (
    OpsgenieRegion.US,
    OpsgenieRegion.EU,
)


# Priorities
class OpsgeniePriority:
    LOW = 1
    MODERATE = 2
    NORMAL = 3
    HIGH = 4
    EMERGENCY = 5


OPSGENIE_PRIORITIES = {
    # Note: This also acts as a reverse lookup mapping
    OpsgeniePriority.LOW: "low",
    OpsgeniePriority.MODERATE: "moderate",
    OpsgeniePriority.NORMAL: "normal",
    OpsgeniePriority.HIGH: "high",
    OpsgeniePriority.EMERGENCY: "emergency",
}

OPSGENIE_PRIORITY_MAP = {
    # Maps against string 'low'
    "l": OpsgeniePriority.LOW,
    # Maps against string 'moderate'
    "m": OpsgeniePriority.MODERATE,
    # Maps against string 'normal'
    "n": OpsgeniePriority.NORMAL,
    # Maps against string 'high'
    "h": OpsgeniePriority.HIGH,
    # Maps against string 'emergency'
    "e": OpsgeniePriority.EMERGENCY,
    # Entries to additionally support (so more like Opsgenie's API)
    "1": OpsgeniePriority.LOW,
    "2": OpsgeniePriority.MODERATE,
    "3": OpsgeniePriority.NORMAL,
    "4": OpsgeniePriority.HIGH,
    "5": OpsgeniePriority.EMERGENCY,
    # Support p-prefix
    "p1": OpsgeniePriority.LOW,
    "p2": OpsgeniePriority.MODERATE,
    "p3": OpsgeniePriority.NORMAL,
    "p4": OpsgeniePriority.HIGH,
    "p5": OpsgeniePriority.EMERGENCY,
}


class NotifyOpsgenie(NotifyBase):
    """A wrapper for Opsgenie Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Opsgenie"

    # The services URL
    service_url = "https://opsgenie.com/"

    # All notification requests are secure
    secure_protocol = "opsgenie"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_opsgenie"

    # The maximum length of the body
    body_maxlen = 15000

    # Our default is to no not use persistent storage beyond in-memory
    # reference
    storage_mode = PersistentStoreMode.AUTO

    # If we don't have the specified min length, then we don't bother using
    # the body directive
    opsgenie_body_minlen = 130

    # The default region to use if one isn't otherwise specified
    opsgenie_default_region = OpsgenieRegion.US

    # The maximum allowable targets within a notification
    default_batch_size = 50

    # Defines our default message mapping
    opsgenie_message_map = {
        # Add a note to existing alert
        NotifyType.INFO: OpsgenieAlertAction.NOTE,
        # Close existing alert
        NotifyType.SUCCESS: OpsgenieAlertAction.CLOSE,
        # Create notice
        NotifyType.WARNING: OpsgenieAlertAction.NEW,
        # Create notice
        NotifyType.FAILURE: OpsgenieAlertAction.NEW,
    }

    # Define object templates
    templates = (
        "{schema}://{apikey}",
        "{schema}://{user}@{apikey}",
        "{schema}://{apikey}/{targets}",
        "{schema}://{user}@{apikey}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "target_escalation": {
                "name": _("Target Escalation"),
                "prefix": "^",
                "type": "string",
                "map_to": "targets",
            },
            "target_schedule": {
                "name": _("Target Schedule"),
                "type": "string",
                "prefix": "*",
                "map_to": "targets",
            },
            "target_user": {
                "name": _("Target User"),
                "type": "string",
                "prefix": "@",
                "map_to": "targets",
            },
            "target_team": {
                "name": _("Target Team"),
                "type": "string",
                "prefix": "#",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets "),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "region": {
                "name": _("Region Name"),
                "type": "choice:string",
                "values": OPSGENIE_REGIONS,
                "default": OpsgenieRegion.US,
                "map_to": "region_name",
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
            "priority": {
                "name": _("Priority"),
                "type": "choice:int",
                "values": OPSGENIE_PRIORITIES,
                "default": OpsgeniePriority.NORMAL,
            },
            "entity": {
                "name": _("Entity"),
                "type": "string",
            },
            "alias": {
                "name": _("Alias"),
                "type": "string",
            },
            "tags": {
                "name": _("Tags"),
                "type": "string",
            },
            "to": {
                "alias_of": "targets",
            },
            "action": {
                "name": _("Action"),
                "type": "choice:string",
                "values": OPSGENIE_ACTIONS,
                "default": OPSGENIE_ACTIONS[0],
            },
        },
    )

    # Map of key-value pairs to use as custom properties of the alert.
    template_kwargs = {
        "details": {
            "name": _("Details"),
            "prefix": "+",
        },
        "mapping": {
            "name": _("Action Mapping"),
            "prefix": ":",
        },
    }

    def __init__(
        self,
        apikey,
        targets,
        region_name=None,
        details=None,
        priority=None,
        alias=None,
        entity=None,
        batch=False,
        tags=None,
        action=None,
        mapping=None,
        **kwargs,
    ):
        """Initialize Opsgenie Object."""
        super().__init__(**kwargs)

        # API Key (associated with project)
        self.apikey = validate_regex(apikey)
        if not self.apikey:
            msg = f"An invalid Opsgenie API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # The Priority of the message
        self.priority = (
            NotifyOpsgenie.template_args["priority"]["default"]
            if not priority
            else next(
                (
                    v
                    for k, v in OPSGENIE_PRIORITY_MAP.items()
                    if str(priority).lower().startswith(k)
                ),
                NotifyOpsgenie.template_args["priority"]["default"],
            )
        )

        # Store our region
        try:
            self.region_name = (
                self.opsgenie_default_region
                if region_name is None
                else region_name.lower()
            )

            if self.region_name not in OPSGENIE_REGIONS:
                # allow the outer except to handle this common response
                raise
        except:
            # Invalid region specified
            msg = f"The Opsgenie region specified ({region_name}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg) from None

        if action and isinstance(action, str):
            self.action = next(
                (a for a in OPSGENIE_ACTIONS if a.startswith(action)), None
            )
            if self.action not in OPSGENIE_ACTIONS:
                msg = f"The Opsgenie action specified ({action}) is invalid."
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            self.action = self.template_args["action"]["default"]

        # Store our mappings
        self.mapping = self.opsgenie_message_map.copy()
        if mapping and isinstance(mapping, dict):
            for _k, _v in mapping.items():
                # Get our mapping
                k = next((t for t in NOTIFY_TYPES if t.startswith(_k)), None)
                if not k:
                    msg = (
                        f"The Opsgenie mapping key specified ({_k}) "
                        "is invalid."
                    )
                    self.logger.warning(msg)
                    raise TypeError(msg)

                _v_lower = _v.lower()
                v = next(
                    (
                        v
                        for v in OPSGENIE_ACTIONS[1:]
                        if v.startswith(_v_lower)
                    ),
                    None,
                )
                if not v:
                    msg = (
                        f"The Opsgenie mapping value (assigned to {k}) "
                        f"specified ({_v}) is invalid."
                    )
                    self.logger.warning(msg)
                    raise TypeError(msg)

                # Update our mapping
                self.mapping[k] = v

        self.details = {}
        if details:
            # Store our extra details
            self.details.update(details)

        # Prepare Batch Mode Flag
        self.batch_size = self.default_batch_size if batch else 1

        # Assign our tags (if defined)
        self.__tags = parse_list(tags)

        # Assign our entity (if defined)
        self.entity = entity

        # Assign our alias (if defined)
        self.alias = alias

        # Initialize our Targets
        self.targets = []

        # Sort our targets
        for _target in parse_list(targets):
            target = _target.strip()
            if len(target) < 2:
                self.logger.debug(f"Ignoring Opsgenie Entry: {target}")
                continue

            if target.startswith(
                NotifyOpsgenie.template_tokens["target_team"]["prefix"]
            ):

                self.targets.append(
                    {"type": OpsgenieCategory.TEAM, "id": target[1:]}
                    if is_uuid(target[1:])
                    else {"type": OpsgenieCategory.TEAM, "name": target[1:]}
                )

            elif target.startswith(
                NotifyOpsgenie.template_tokens["target_schedule"]["prefix"]
            ):

                self.targets.append(
                    {"type": OpsgenieCategory.SCHEDULE, "id": target[1:]}
                    if is_uuid(target[1:])
                    else {
                        "type": OpsgenieCategory.SCHEDULE,
                        "name": target[1:],
                    }
                )

            elif target.startswith(
                NotifyOpsgenie.template_tokens["target_escalation"]["prefix"]
            ):

                self.targets.append(
                    {"type": OpsgenieCategory.ESCALATION, "id": target[1:]}
                    if is_uuid(target[1:])
                    else {
                        "type": OpsgenieCategory.ESCALATION,
                        "name": target[1:],
                    }
                )

            elif target.startswith(
                NotifyOpsgenie.template_tokens["target_user"]["prefix"]
            ):

                self.targets.append(
                    {"type": OpsgenieCategory.USER, "id": target[1:]}
                    if is_uuid(target[1:])
                    else {
                        "type": OpsgenieCategory.USER,
                        "username": target[1:],
                    }
                )

            else:
                # Ambiguious entry; treat it as a user but not before
                # displaying a warning to the end user first:
                self.logger.debug(
                    "Treating ambigious Opsgenie target %s as a user", target
                )
                self.targets.append(
                    {"type": OpsgenieCategory.USER, "id": target}
                    if is_uuid(target)
                    else {"type": OpsgenieCategory.USER, "username": target}
                )

    def _fetch(self, method, url, payload, params=None):
        """Performs server retrieval/update and returns JSON Response."""
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Authorization": f"GenieKey {self.apikey}",
        }

        # Some Debug Logging
        self.logger.debug(
            f"Opsgenie POST URL: {url} (cert_verify={self.verify_certificate})"
        )
        self.logger.debug(f"Opsgenie Payload: {payload}")

        # Initialize our response object
        content = {}

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = method(
                url,
                data=dumps(payload),
                params=params,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            # A Response might look like:
            # {
            #     "result": "Request will be processed",
            #     "took": 0.302,
            #     "requestId": "43a29c5c-3dbf-4fa4-9c26-f4f71023e120"
            # }

            try:
                # Update our response object
                content = loads(r.content)

            except (AttributeError, TypeError, ValueError):
                # ValueError = r.content is Unparsable
                # TypeError = r.content is None
                # AttributeError = r is None
                content = {}

            if r.status_code not in (
                requests.codes.accepted,
                requests.codes.ok,
            ):
                status_str = NotifyBase.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Opsgenie notification:"
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                return (False, content.get("requestId"))

            # If we reach here; the message was sent
            self.logger.info("Sent Opsgenie notification")
            self.logger.debug(f"Response Details:\r\n{r.content}")

            return (True, content.get("requestId"))

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Opsgenie notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

        return (False, content.get("requestId"))

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Opsgenie Notification."""

        # Get our Opsgenie Action
        action = (
            OPSGENIE_ALERT_MAP[notify_type]
            if self.action == OpsgenieAlertAction.MAP
            else self.action
        )

        # Prepare our URL as it's based on our hostname
        notify_url = OPSGENIE_API_LOOKUP[self.region_name]

        # Initialize our has_error flag
        has_error = False

        # Default method is to post
        method = requests.post

        # For indexing in persistent store
        key = hashlib.sha1(
            (
                self.entity
                if self.entity
                else (
                    self.alias
                    if self.alias
                    else (title if title else self.app_id)
                )
            ).encode("utf-8")
        ).hexdigest()[0:10]

        # Get our Opsgenie Request IDs
        request_ids = self.store.get(key, [])
        if not isinstance(request_ids, list):
            request_ids = []

        if action == OpsgenieAlertAction.NEW:
            # Create a copy ouf our details object
            details = self.details.copy()
            if "type" not in details:
                details["type"] = notify_type.value

            # Use body if title not set
            title_body = title if title else body

            # Prepare our payload
            payload = {
                "source": self.app_desc,
                "message": title_body,
                "description": body,
                "details": details,
                "priority": f"P{self.priority}",
            }

            # Use our body directive if we exceed the minimum message
            # limitation
            if len(payload["message"]) > self.opsgenie_body_minlen:
                payload["message"] = (
                    f"{title_body[:self.opsgenie_body_minlen - 3]}..."
                )

            if self.__tags:
                payload["tags"] = self.__tags

            if self.entity:
                payload["entity"] = self.entity

            if self.alias:
                payload["alias"] = self.alias

            if self.user:
                payload["user"] = self.user

            # reset our request IDs - we will re-populate them
            request_ids = []

            length = len(self.targets) if self.targets else 1
            for index in range(0, length, self.batch_size):
                if self.targets:
                    # If there were no targets identified, then we simply
                    # just iterate once without the responders set
                    payload["responders"] = self.targets[
                        index : index + self.batch_size
                    ]

                # Perform our post
                success, request_id = self._fetch(method, notify_url, payload)

                if success and request_id:
                    # Save our response
                    request_ids.append(request_id)

                else:
                    has_error = True

            # Store our entries for a maximum of 60 days
            self.store.set(key, request_ids, expires=60 * 60 * 24 * 60)

        elif request_ids:
            # Prepare our payload
            payload = {
                "source": self.app_desc,
                "note": body,
            }

            if self.user:
                payload["user"] = self.user

            # Prepare our Identifier type
            params = {
                "identifierType": "id",
            }

            for request_id in request_ids:
                if action == OpsgenieAlertAction.DELETE:
                    # Update our URL
                    url = f"{notify_url}/{request_id}"
                    method = requests.delete

                elif action == OpsgenieAlertAction.ACKNOWLEDGE:
                    url = f"{notify_url}/{request_id}/acknowledge"

                elif action == OpsgenieAlertAction.CLOSE:
                    url = f"{notify_url}/{request_id}/close"

                else:  # action == OpsgenieAlertAction.CLOSE:
                    url = f"{notify_url}/{request_id}/notes"

                # Perform our post
                success, _ = self._fetch(method, url, payload, params)

                if not success:
                    has_error = True

            if not has_error and action == OpsgenieAlertAction.DELETE:
                # Remove cached entry
                self.store.clear(key)

        else:
            self.logger.info(
                "No Opsgenie notification sent due to (nothing to %s) "
                "condition",
                self.action,
            )

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.region_name, self.apikey)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "action": self.action,
            "region": self.region_name,
            "priority": (
                OPSGENIE_PRIORITIES[self.template_args["priority"]["default"]]
                if self.priority not in OPSGENIE_PRIORITIES
                else OPSGENIE_PRIORITIES[self.priority]
            ),
            "batch": "yes" if self.batch_size > 1 else "no",
        }

        # Assign our entity value (if defined)
        if self.entity:
            params["entity"] = self.entity

        # Assign our alias value (if defined)
        if self.alias:
            params["alias"] = self.alias

        # Assign our tags (if specifed)
        if self.__tags:
            params["tags"] = ",".join(self.__tags)

        # Append our details into our parameters
        params.update({f"+{k}": v for k, v in self.details.items()})

        # Append our assignment extra's into our parameters
        params.update({f":{k.value}": v for k, v in self.mapping.items()})

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # A map allows us to map our target types so they can be correctly
        # placed back into your URL below. Hence map the 'user' -> '@'
        __map = {
            OpsgenieCategory.USER: NotifyOpsgenie.template_tokens[
                "target_user"
            ]["prefix"],
            OpsgenieCategory.SCHEDULE: NotifyOpsgenie.template_tokens[
                "target_schedule"
            ]["prefix"],
            OpsgenieCategory.ESCALATION: NotifyOpsgenie.template_tokens[
                "target_escalation"
            ]["prefix"],
            OpsgenieCategory.TEAM: NotifyOpsgenie.template_tokens[
                "target_team"
            ]["prefix"],
        }

        return "{schema}://{user}{apikey}/{targets}?{params}".format(
            schema=self.secure_protocol,
            user=f"{self.user}@" if self.user else "",
            apikey=self.pprint(self.apikey, privacy, safe=""),
            targets="/".join([
                NotifyOpsgenie.quote(
                    "{}{}".format(
                        __map[x["type"]],
                        x.get("id", x.get("name", x.get("username"))),
                    )
                )
                for x in self.targets
            ]),
            params=NotifyOpsgenie.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        #
        # Factor batch into calculation
        #
        targets = len(self.targets)
        if self.batch_size > 1:
            targets = int(targets / self.batch_size) + (
                1 if targets % self.batch_size else 0
            )

        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The API Key is stored in the hostname
        results["apikey"] = NotifyOpsgenie.unquote(results["host"])

        # Get our Targets
        results["targets"] = NotifyOpsgenie.split_path(results["fullpath"])

        # Add our Meta Detail keys
        results["details"] = {
            NotifyBase.unquote(x): NotifyBase.unquote(y)
            for x, y in results["qsd+"].items()
        }

        # Set our priority
        if "priority" in results["qsd"] and len(results["qsd"]["priority"]):
            results["priority"] = NotifyOpsgenie.unquote(
                results["qsd"]["priority"]
            )

        # Get Batch Boolean (if set)
        results["batch"] = parse_bool(
            results["qsd"].get(
                "batch", NotifyOpsgenie.template_args["batch"]["default"]
            )
        )

        if "apikey" in results["qsd"] and len(results["qsd"]["apikey"]):
            results["apikey"] = NotifyOpsgenie.unquote(
                results["qsd"]["apikey"]
            )

        if "tags" in results["qsd"] and len(results["qsd"]["tags"]):
            # Extract our tags
            results["tags"] = parse_list(
                NotifyOpsgenie.unquote(results["qsd"]["tags"])
            )

        if "region" in results["qsd"] and len(results["qsd"]["region"]):
            # Extract our region
            results["region_name"] = NotifyOpsgenie.unquote(
                results["qsd"]["region"]
            )

        if "entity" in results["qsd"] and len(results["qsd"]["entity"]):
            # Extract optional entity field
            results["entity"] = NotifyOpsgenie.unquote(
                results["qsd"]["entity"]
            )

        if "alias" in results["qsd"] and len(results["qsd"]["alias"]):
            # Extract optional alias field
            results["alias"] = NotifyOpsgenie.unquote(results["qsd"]["alias"])

        # Handle 'to' email address
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"].append(results["qsd"]["to"])

        # Store our action (if defined)
        if "action" in results["qsd"] and len(results["qsd"]["action"]):
            results["action"] = NotifyOpsgenie.unquote(
                results["qsd"]["action"]
            )

        # store any custom mapping defined
        results["mapping"] = {
            NotifyOpsgenie.unquote(x): NotifyOpsgenie.unquote(y)
            for x, y in results["qsd:"].items()
        }

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import re

import requests

from ..common import ConfigFormat, ContentIncludeMode
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from .base import ConfigBase

# Support YAML formats
# text/yaml
# text/x-yaml
# application/yaml
# application/x-yaml
MIME_IS_YAML = re.compile("(text|application)/(x-)?yaml", re.I)

# Support TEXT formats
# text/plain
# text/html
MIME_IS_TEXT = re.compile("text/(plain|html)", re.I)


class ConfigHTTP(ConfigBase):
    """A wrapper for HTTP based configuration sources."""

    # The default descriptive name associated with the service
    service_name = _("Web Based")

    # The default protocol
    protocol = "http"

    # The default secure protocol
    secure_protocol = "https"

    # If an HTTP error occurs, define the number of characters you still want
    # to read back.  This is useful for debugging purposes, but nothing else.
    # The idea behind enforcing this kind of restriction is to prevent abuse
    # from queries to services that may be untrusted.
    max_error_buffer_size = 2048

    # Configuration file inclusion can always include this type
    allow_cross_includes = ContentIncludeMode.ALWAYS

    def __init__(self, headers=None, **kwargs):
        """Initialize HTTP Object.

        headers can be a dictionary of key/value pairs that you want to
        additionally include as part of the server headers to post with
        """
        super().__init__(**kwargs)

        self.schema = "https" if self.secure else "http"

        self.fullpath = kwargs.get("fullpath")
        if not isinstance(self.fullpath, str):
            self.fullpath = "/"

        self.headers = {}
        if headers:
            # Store our extra headers
            self.headers.update(headers)

        return

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Prepare our cache value
        if isinstance(self.cache, bool) or not self.cache:
            cache = "yes" if self.cache else "no"

        else:
            cache = int(self.cache)

        # Define any arguments set
        params = {
            "encoding": self.encoding,
            "cache": cache,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        if self.config_format:
            # A format was enforced; make sure it's passed back with the url
            params["format"] = self.config_format

        # Append our headers into our args
        params.update({f"+{k}": v for k, v in self.headers.items()})

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=self.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=self.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else 80
        return "{schema}://{auth}{hostname}{port}{fullpath}/?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            hostname=self.quote(self.host, safe=""),
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            fullpath=self.quote(self.fullpath, safe="/"),
            params=self.urlencode(params),
        )

    def read(self, **kwargs):
        """Perform retrieval of the configuration based on the specified
        request."""

        # prepare XML Object
        headers = {
            "User-Agent": self.app_id,
        }

        # Apply any/all header over-rides defined
        headers.update(self.headers)

        auth = None
        if self.user:
            auth = (self.user, self.password)

        url = f"{self.schema}://{self.host}"
        if isinstance(self.port, int):
            url += f":{self.port}"

        url += self.fullpath

        self.logger.debug(
            f"HTTP POST URL: {url} (cert_verify={self.verify_certificate!r})"
        )

        # Prepare our response object
        response = None

        # Where our request object will temporarily live.
        r = None

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            # Make our request
            with requests.post(
                url,
                headers=headers,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
                stream=True,
            ) as r:

                # Handle Errors
                r.raise_for_status()

                # Get our file-size (if known)
                try:
                    file_size = int(r.headers.get("Content-Length", "0"))
                except (TypeError, ValueError):
                    # Handle edge case where Content-Length is a bad value
                    file_size = 0

                # Store our response
                if (
                    self.max_buffer_size > 0
                    and file_size > self.max_buffer_size
                ):

                    # Provide warning of data truncation
                    self.logger.error(
                        "HTTP config response exceeds maximum buffer length "
                        f"({int(self.max_buffer_size / 1024)}KB);"
                    )

                    # Return None - buffer execeeded
                    return None

                # Store our result (but no more than our buffer length)
                response = r.text[: self.max_buffer_size + 1]

                # Verify that our content did not exceed the buffer size:
                if len(response) > self.max_buffer_size:
                    # Provide warning of data truncation
                    self.logger.error(
                        "HTTP config response exceeds maximum buffer length "
                        f"({int(self.max_buffer_size / 1024)}KB);"
                    )

                    # Return None - buffer execeeded
                    return None

                # Detect config format based on mime if the format isn't
                # already enforced
                content_type = r.headers.get(
                    "Content-Type", "application/octet-stream"
                )
                if self.config_format is None and content_type:
                    if MIME_IS_YAML.match(content_type) is not None:

                        # YAML data detected based on header content
                        self.default_config_format = ConfigFormat.YAML

                    elif MIME_IS_TEXT.match(content_type) is not None:

                        # TEXT data detected based on header content
                        self.default_config_format = ConfigFormat.TEXT

        except requests.RequestException as e:
            self.logger.error(
                "A Connection error occurred retrieving HTTP "
                f"configuration from {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return None (signifying a failure)
            return None

        # Return our response object
        return response

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = ConfigBase.parse_url(url)

        if not results:
            # We're done early as we couldn't load the results
            return results

        # Add our headers that the user can potentially over-ride if they wish
        # to to our returned result set
        results["headers"] = results["qsd-"]
        results["headers"].update(results["qsd+"])

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Create an account https://messagebird.com if you don't already have one
#
# Get your (apikey) and api example from the dashboard here:
#   - https://dashboard.messagebird.com/en/user/index
#

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import is_phone_no, parse_phone_no, validate_regex
from .base import NotifyBase


class NotifyMessageBird(NotifyBase):
    """A wrapper for MessageBird Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "MessageBird"

    # The services URL
    service_url = "https://messagebird.com"

    # The default protocol
    secure_protocol = "msgbird"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_messagebird"

    # MessageBird uses the http protocol with JSON requests
    notify_url = "https://rest.messagebird.com/messages"

    # The maximum length of the body
    body_maxlen = 160

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = (
        "{schema}://{apikey}/{source}",
        "{schema}://{apikey}/{source}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[a-z0-9]{25}$", "i"),
            },
            "source": {
                "name": _("Source Phone No"),
                "type": "string",
                "prefix": "+",
                "required": True,
                "regex": (r"^[0-9\s)(+-]+$", "i"),
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "source",
            },
        },
    )

    def __init__(self, apikey, source, targets=None, **kwargs):
        """Initialize MessageBird Object."""
        super().__init__(**kwargs)

        # API Key (associated with project)
        self.apikey = validate_regex(
            apikey, *self.template_tokens["apikey"]["regex"]
        )
        if not self.apikey:
            msg = f"An invalid MessageBird API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        result = is_phone_no(source)
        if not result:
            msg = f"The MessageBird source specified ({source}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our source
        self.source = result["full"]

        # Parse our targets
        self.targets = []

        targets = parse_phone_no(targets)
        if not targets:
            # No sources specified, use our own phone no
            self.targets.append(self.source)
            return

        # otherwise, store all of our target numbers
        for target in targets:
            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
                continue

            # store valid phone number
            self.targets.append(result["full"])

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform MessageBird Notification."""

        if len(self.targets) == 0:
            # There were no services to notify
            self.logger.warning("There were no MessageBird targets to notify.")
            return False

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": f"AccessKey {self.apikey}",
        }

        # Prepare our payload
        payload = {
            "originator": f"+{self.source}",
            "recipients": None,
            "body": body,
        }

        # Create a copy of the targets list
        targets = list(self.targets)

        while len(targets):
            # Get our target to notify
            target = targets.pop(0)

            # Prepare our user
            payload["recipients"] = f"+{target}"

            # Some Debug Logging
            self.logger.debug(
                "MessageBird POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"MessageBird Payload: {payload}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                # Sample output of a successful transmission
                # {
                #   "originator": "+15553338888",
                #   "body": "test",
                #   "direction": "mt",
                #   "mclass": 1,
                #   "reference": null,
                #   "createdDatetime": "2019-08-22T01:32:18+00:00",
                #   "recipients": {
                #     "totalCount": 1,
                #     "totalSentCount": 1,
                #     "totalDeliveredCount": 0,
                #     "totalDeliveryFailedCount": 0,
                #     "items": [
                #       {
                #         "status": "sent",
                #         "statusDatetime": "2019-08-22T01:32:18+00:00",
                #         "recipient": 15553338888,
                #         "messagePartCount": 1
                #       }
                #     ]
                #   },
                #   "validity": null,
                #   "gateway": 10,
                #   "typeDetails": {},
                #   "href": "https://rest.messagebird.com/messages/\
                #       b5d424244a5b4fd0b5b5728bccaafc23",
                #   "datacoding": "plain",
                #   "scheduledDatetime": null,
                #   "type": "sms",
                #   "id": "b5d424244a5b4fd0b5b5728bccaafc23"
                # }

                if r.status_code not in (
                    requests.codes.ok,
                    requests.codes.created,
                ):
                    # We had a problem
                    status_str = NotifyMessageBird.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send MessageBird notification to {}: "
                        "{}{}error={}.".format(
                            ",".join(target),
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        f"Sent MessageBird notification to {target}."
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending"
                    f" MessageBird:{target} "
                    + "notification."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.apikey, self.source)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return "{schema}://{apikey}/{source}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            apikey=self.pprint(self.apikey, privacy, safe=""),
            source=self.source,
            targets="/".join(
                [NotifyMessageBird.quote(x, safe="") for x in self.targets]
            ),
            params=NotifyMessageBird.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifyMessageBird.split_path(results["fullpath"])

        try:
            # The first path entry is the source/originator
            results["source"] = results["targets"].pop(0)

        except IndexError:
            # No path specified... this URL is potentially un-parseable; we can
            # hope for a from= entry
            results["source"] = None

        # The hostname is our authentication key
        results["apikey"] = NotifyMessageBird.unquote(results["host"])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyMessageBird.parse_phone_no(
                results["qsd"]["to"]
            )

        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifyMessageBird.unquote(
                results["qsd"]["from"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import socket

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_bool
from .base import NotifyBase


class syslog:
    """Extrapoloated information from the syslog library so that this plugin
    would not be dependent on it."""

    # Notification Categories
    LOG_KERN = 0
    LOG_USER = 8
    LOG_MAIL = 16
    LOG_DAEMON = 24
    LOG_AUTH = 32
    LOG_SYSLOG = 40
    LOG_LPR = 48
    LOG_NEWS = 56
    LOG_UUCP = 64
    LOG_CRON = 72
    LOG_LOCAL0 = 128
    LOG_LOCAL1 = 136
    LOG_LOCAL2 = 144
    LOG_LOCAL3 = 152
    LOG_LOCAL4 = 160
    LOG_LOCAL5 = 168
    LOG_LOCAL6 = 176
    LOG_LOCAL7 = 184

    # Notification Types
    LOG_INFO = 6
    LOG_NOTICE = 5
    LOG_WARNING = 4
    LOG_CRIT = 2


class SyslogFacility:
    """All of the supported facilities."""

    KERN = "kern"
    USER = "user"
    MAIL = "mail"
    DAEMON = "daemon"
    AUTH = "auth"
    SYSLOG = "syslog"
    LPR = "lpr"
    NEWS = "news"
    UUCP = "uucp"
    CRON = "cron"
    LOCAL0 = "local0"
    LOCAL1 = "local1"
    LOCAL2 = "local2"
    LOCAL3 = "local3"
    LOCAL4 = "local4"
    LOCAL5 = "local5"
    LOCAL6 = "local6"
    LOCAL7 = "local7"


SYSLOG_FACILITY_MAP = {
    SyslogFacility.KERN: syslog.LOG_KERN,
    SyslogFacility.USER: syslog.LOG_USER,
    SyslogFacility.MAIL: syslog.LOG_MAIL,
    SyslogFacility.DAEMON: syslog.LOG_DAEMON,
    SyslogFacility.AUTH: syslog.LOG_AUTH,
    SyslogFacility.SYSLOG: syslog.LOG_SYSLOG,
    SyslogFacility.LPR: syslog.LOG_LPR,
    SyslogFacility.NEWS: syslog.LOG_NEWS,
    SyslogFacility.UUCP: syslog.LOG_UUCP,
    SyslogFacility.CRON: syslog.LOG_CRON,
    SyslogFacility.LOCAL0: syslog.LOG_LOCAL0,
    SyslogFacility.LOCAL1: syslog.LOG_LOCAL1,
    SyslogFacility.LOCAL2: syslog.LOG_LOCAL2,
    SyslogFacility.LOCAL3: syslog.LOG_LOCAL3,
    SyslogFacility.LOCAL4: syslog.LOG_LOCAL4,
    SyslogFacility.LOCAL5: syslog.LOG_LOCAL5,
    SyslogFacility.LOCAL6: syslog.LOG_LOCAL6,
    SyslogFacility.LOCAL7: syslog.LOG_LOCAL7,
}

SYSLOG_FACILITY_RMAP = {
    syslog.LOG_KERN: SyslogFacility.KERN,
    syslog.LOG_USER: SyslogFacility.USER,
    syslog.LOG_MAIL: SyslogFacility.MAIL,
    syslog.LOG_DAEMON: SyslogFacility.DAEMON,
    syslog.LOG_AUTH: SyslogFacility.AUTH,
    syslog.LOG_SYSLOG: SyslogFacility.SYSLOG,
    syslog.LOG_LPR: SyslogFacility.LPR,
    syslog.LOG_NEWS: SyslogFacility.NEWS,
    syslog.LOG_UUCP: SyslogFacility.UUCP,
    syslog.LOG_CRON: SyslogFacility.CRON,
    syslog.LOG_LOCAL0: SyslogFacility.LOCAL0,
    syslog.LOG_LOCAL1: SyslogFacility.LOCAL1,
    syslog.LOG_LOCAL2: SyslogFacility.LOCAL2,
    syslog.LOG_LOCAL3: SyslogFacility.LOCAL3,
    syslog.LOG_LOCAL4: SyslogFacility.LOCAL4,
    syslog.LOG_LOCAL5: SyslogFacility.LOCAL5,
    syslog.LOG_LOCAL6: SyslogFacility.LOCAL6,
    syslog.LOG_LOCAL7: SyslogFacility.LOCAL7,
}

# Used as a lookup when handling the Apprise -> Syslog Mapping
SYSLOG_PUBLISH_MAP = {
    NotifyType.INFO: syslog.LOG_INFO,
    NotifyType.SUCCESS: syslog.LOG_NOTICE,
    NotifyType.FAILURE: syslog.LOG_CRIT,
    NotifyType.WARNING: syslog.LOG_WARNING,
}


class NotifyRSyslog(NotifyBase):
    """A wrapper for Remote Syslog Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Remote Syslog"

    # The services URL
    service_url = "https://tools.ietf.org/html/rfc5424"

    # The default protocol
    protocol = "rsyslog"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_rsyslog"

    # Disable throttle rate for RSyslog requests
    request_rate_per_sec = 0

    # Define object templates
    templates = (
        "{schema}://{host}",
        "{schema}://{host}:{port}",
        "{schema}://{host}/{facility}",
        "{schema}://{host}:{port}/{facility}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "facility": {
                "name": _("Facility"),
                "type": "choice:string",
                "values": list(SYSLOG_FACILITY_MAP),
                "default": SyslogFacility.USER,
                "required": True,
            },
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
                "default": 514,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "facility": {
                # We map back to the same element defined in template_tokens
                "alias_of": "facility",
            },
            "logpid": {
                "name": _("Log PID"),
                "type": "bool",
                "default": True,
                "map_to": "log_pid",
            },
        },
    )

    def __init__(self, facility=None, log_pid=True, **kwargs):
        """Initialize RSyslog Object."""
        super().__init__(**kwargs)

        if facility:
            try:
                self.facility = SYSLOG_FACILITY_MAP[facility]

            except KeyError:
                msg = f"An invalid syslog facility ({facility}) was specified."
                self.logger.warning(msg)
                raise TypeError(msg) from None

        else:
            self.facility = SYSLOG_FACILITY_MAP[
                self.template_tokens["facility"]["default"]
            ]

        # Include PID with each message.
        self.log_pid = log_pid

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform RSyslog Notification."""

        if title:
            # Format title
            body = f"{title}: {body}"

        # Always call throttle before any remote server i/o is made
        self.throttle()
        host = self.host
        port = (
            self.port if self.port else self.template_tokens["port"]["default"]
        )

        priority = SYSLOG_PUBLISH_MAP[notify_type] + self.facility * 8
        payload = f"<{priority}>- {os.getpid()} {body}" \
            if self.log_pid else f"<{priority}>- {body}"

        # send UDP packet to upstream server
        self.logger.debug(
            "RSyslog Host: %s:%d/%s",
            host,
            port,
            SYSLOG_FACILITY_RMAP[self.facility],
        )
        self.logger.debug(f"RSyslog Payload: {payload!s}")

        # our sent bytes
        sent = 0

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.settimeout(self.socket_connect_timeout)
            sent = sock.sendto(payload.encode("utf-8"), (host, port))
            sock.close()

        except socket.gaierror as e:
            self.logger.warning(
                "A connection error occurred sending RSyslog "
                "notification to %s:%d/%s",
                host,
                port,
                SYSLOG_FACILITY_RMAP[self.facility],
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        except socket.timeout as e:
            self.logger.warning(
                "A connection timeout occurred sending RSyslog "
                "notification to %s:%d/%s",
                host,
                port,
                SYSLOG_FACILITY_RMAP[self.facility],
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        if sent < len(payload):
            self.logger.warning(
                "RSyslog sent %d byte(s) but intended to send %d byte(s)",
                sent,
                len(payload),
            )
            return False

        self.logger.info("Sent RSyslog notification.")

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.protocol,
            self.host,
            (
                self.port
                if self.port
                else self.template_tokens["port"]["default"]
            ),
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "logpid": "yes" if self.log_pid else "no",
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{hostname}{port}/{facility}/?{params}".format(
            schema=self.protocol,
            hostname=NotifyRSyslog.quote(self.host, safe=""),
            port=(
                ""
                if self.port is None
                or self.port == self.template_tokens["port"]["default"]
                else f":{self.port}"
            ),
            facility=(
                self.template_tokens["facility"]["default"]
                if self.facility not in SYSLOG_FACILITY_RMAP
                else SYSLOG_FACILITY_RMAP[self.facility]
            ),
            params=NotifyRSyslog.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        tokens = []

        # Get our path values
        tokens.extend(NotifyRSyslog.split_path(results["fullpath"]))

        # Initialization
        facility = None

        if tokens:
            # Store the last entry as the facility
            facility = tokens[-1].lower()

        # However if specified on the URL, that will over-ride what was
        # identified
        if "facility" in results["qsd"] and len(results["qsd"]["facility"]):
            facility = results["qsd"]["facility"].lower()

        if facility and facility not in SYSLOG_FACILITY_MAP:
            # Find first match; if no match is found we set the result
            # to the matching key.  This allows us to throw a TypeError
            # during the __init__() call. The benifit of doing this
            # check here is if we do have a valid match, we can support
            # short form matches like 'u' which will match against user
            facility = next(
                (f for f in SYSLOG_FACILITY_MAP if f.startswith(facility)),
                facility,
            )

        # Save facility if set
        if facility:
            results["facility"] = facility

        # Include PID as part of the message logged
        results["log_pid"] = parse_bool(
            results["qsd"].get(
                "logpid", NotifyRSyslog.template_args["logpid"]["default"]
            )
        )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this plugin, you must have a Africas Talking Account setup; See here:
#  https://account.africastalking.com/
#  From here... acquire your APIKey
#
# API Details: https://developers.africastalking.com/docs/sms/sending/bulk
import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import (
    is_phone_no,
    parse_bool,
    parse_phone_no,
    validate_regex,
)
from .base import NotifyBase


class AfricasTalkingSMSMode:
    """Africas Talking SMS Mode."""

    # BulkSMS Mode
    BULKSMS = "bulksms"

    # Premium Mode
    PREMIUM = "premium"

    # Sandbox Mode
    SANDBOX = "sandbox"


# Define the types in a list for validation purposes
AFRICAS_TALKING_SMS_MODES = (
    AfricasTalkingSMSMode.BULKSMS,
    AfricasTalkingSMSMode.PREMIUM,
    AfricasTalkingSMSMode.SANDBOX,
)


# Extend HTTP Error Messages
AFRICAS_TALKING_HTTP_ERROR_MAP = {
    100: "Processed",
    101: "Sent",
    102: "Queued",
    401: "Risk Hold",
    402: "Invalid Sender ID",
    403: "Invalid Phone Number",
    404: "Unsupported Number Type",
    405: "Insufficient Balance",
    406: "User In Blacklist",
    407: "Could Not Route",
    409: "Do Not Disturb Rejection",
    500: "Internal Server Error",
    501: "Gateway Error",
    502: "Rejected By Gateway",
}


class NotifyAfricasTalking(NotifyBase):
    """A wrapper for Africas Talking Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Africas Talking"

    # The services URL
    service_url = "https://africastalking.com/"

    # The default secure protocol
    secure_protocol = "atalk"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_africas_talking"

    # Africas Talking API Request URLs
    notify_url = {
        AfricasTalkingSMSMode.BULKSMS: (
            "https://api.africastalking.com/version1/messaging"
        ),
        AfricasTalkingSMSMode.PREMIUM: (
            "https://content.africastalking.com/version1/messaging"
        ),
        AfricasTalkingSMSMode.SANDBOX: (
            "https://api.sandbox.africastalking.com/version1/messaging"
        ),
    }

    # The maximum allowable characters allowed in the title per message
    title_maxlen = 0

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 160

    # The maximum amount of phone numbers that can reside within a single
    # batch transfer
    default_batch_size = 50

    # Define object templates
    templates = ("{schema}://{appuser}@{apikey}/{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "appuser": {
                "name": _("App User Name"),
                "type": "string",
                "regex": (r"^[A-Z0-9_-]+$", "i"),
                "required": True,
            },
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[A-Z0-9_-]+$", "i"),
            },
            "target_phone": {
                "name": _("Target Phone"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "apikey": {
                "alias_of": "apikey",
            },
            "from": {
                # Your registered short code or alphanumeric
                "name": _("From"),
                "type": "string",
                "default": "AFRICASTKNG",
                "map_to": "sender",
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
            "mode": {
                "name": _("SMS Mode"),
                "type": "choice:string",
                "values": AFRICAS_TALKING_SMS_MODES,
                "default": AFRICAS_TALKING_SMS_MODES[0],
            },
        },
    )

    def __init__(
        self,
        appuser,
        apikey,
        targets=None,
        sender=None,
        batch=None,
        mode=None,
        **kwargs,
    ):
        """Initialize Africas Talking Object."""
        super().__init__(**kwargs)

        self.appuser = validate_regex(
            appuser, *self.template_tokens["appuser"]["regex"]
        )
        if not self.appuser:
            msg = (
                f"The Africas Talking appuser specified ({appuser}) is"
                " invalid."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        self.apikey = validate_regex(
            apikey, *self.template_tokens["apikey"]["regex"]
        )
        if not self.apikey:
            msg = (
                f"The Africas Talking apikey specified ({apikey}) is invalid."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Prepare Sender
        self.sender = (
            self.template_args["from"]["default"] if sender is None else sender
        )

        # Prepare Batch Mode Flag
        self.batch = (
            self.template_args["batch"]["default"] if batch is None else batch
        )

        self.mode = (
            self.template_args["mode"]["default"]
            if not isinstance(mode, str)
            else mode.lower()
        )

        if isinstance(mode, str) and mode:
            self.mode = next(
                (
                    a
                    for a in AFRICAS_TALKING_SMS_MODES
                    if a.startswith(mode.lower())
                ),
                None,
            )

            if self.mode not in AFRICAS_TALKING_SMS_MODES:
                msg = (
                    f"The Africas Talking mode specified ({mode}) is invalid."
                )
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            self.mode = self.template_args["mode"]["default"]

        # Parse our targets
        self.targets = []

        for target in parse_phone_no(targets):
            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
                continue

            # store valid phone number
            # Carry forward '+' if defined, otherwise do not...
            self.targets.append(
                ("+" + result["full"])
                if target.lstrip()[0] == "+"
                else result["full"]
            )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Africas Talking Notification."""

        if not self.targets:
            # There is no one to email; we're done
            self.logger.warning(
                "There are no Africas Talking recipients to notify"
            )
            return False

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
            "apiKey": self.apikey,
        }

        # error tracking (used for function return)
        has_error = False

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        # Create a copy of the target list
        for index in range(0, len(self.targets), batch_size):
            # Prepare our payload
            payload = {
                "username": self.appuser,
                "to": ",".join(self.targets[index : index + batch_size]),
                "from": self.sender,
                "message": body,
            }

            # Acquire our URL
            notify_url = self.notify_url[self.mode]

            self.logger.debug(
                "Africas Talking POST URL:"
                f" {notify_url} (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"Africas Talking Payload: {payload!s}")

            # Printable target detail
            p_target = (
                self.targets[index]
                if batch_size == 1
                else f"{len(self.targets[index:index + batch_size])} target(s)"
            )

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    notify_url,
                    data=payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                # Sample response
                # {
                #     "SMSMessageData": {
                #         "Message": "Sent to 1/1 Total Cost: KES 0.8000",
                #         "Recipients": [{
                #             "statusCode": 101,
                #             "number": "+254711XXXYYY",
                #             "status": "Success",
                #             "cost": "KES 0.8000",
                #             "messageId": "ATPid_SampleTxnId123"
                #         }]
                #     }
                # }

                if r.status_code not in (100, 101, 102, requests.codes.ok):
                    # We had a problem
                    status_str = (
                        NotifyAfricasTalking.http_response_code_lookup(
                            r.status_code, AFRICAS_TALKING_HTTP_ERROR_MAP
                        )
                    )

                    self.logger.warning(
                        "Failed to send Africas Talking notification to {}: "
                        "{}{}error={}.".format(
                            p_target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        f"Sent Africas Talking notification to {p_target}."
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Africas Talking "
                    f"notification to {p_target}."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.appuser, self.apikey)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "batch": "yes" if self.batch else "no",
        }

        if self.sender != self.template_args["from"]["default"]:
            # Set our sender if it was set
            params["from"] = self.sender

        if self.mode != self.template_args["mode"]["default"]:
            # Set our mode
            params["mode"] = self.mode

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{appuser}@{apikey}/{targets}?{params}".format(
            schema=self.secure_protocol,
            appuser=NotifyAfricasTalking.quote(self.appuser, safe=""),
            apikey=self.pprint(self.apikey, privacy, safe=""),
            targets="/".join(
                [NotifyAfricasTalking.quote(x, safe="+") for x in self.targets]
            ),
            params=NotifyAfricasTalking.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        #
        # Factor batch into calculation
        #
        batch_size = 1 if not self.batch else self.default_batch_size
        targets = len(self.targets)
        if batch_size > 1:
            targets = int(targets / batch_size) + (
                1 if targets % batch_size else 0
            )

        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The Application User ID
        results["appuser"] = NotifyAfricasTalking.unquote(results["user"])

        # Prepare our targets
        results["targets"] = []

        # Our Application APIKey
        if "apikey" in results["qsd"] and len(results["qsd"]["apikey"]):
            # Store our apikey if specified as keyword
            results["apikey"] = NotifyAfricasTalking.unquote(
                results["qsd"]["apikey"]
            )

            # This means our host is actually a phone number (target)
            results["targets"].append(
                NotifyAfricasTalking.unquote(results["host"])
            )

        else:
            # First item is our apikey
            results["apikey"] = NotifyAfricasTalking.unquote(results["host"])

        # Store our remaining targets found on path
        results["targets"].extend(
            NotifyAfricasTalking.split_path(results["fullpath"])
        )

        # The 'from' makes it easier to use yaml configuration
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["sender"] = NotifyAfricasTalking.unquote(
                results["qsd"]["from"]
            )

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyAfricasTalking.parse_phone_no(
                results["qsd"]["to"]
            )

        # Get our Mode
        if "mode" in results["qsd"] and len(results["qsd"]["mode"]):
            results["mode"] = NotifyAfricasTalking.unquote(
                results["qsd"]["mode"]
            )

        # Get Batch Mode Flag
        results["batch"] = parse_bool(
            results["qsd"].get(
                "batch", NotifyAfricasTalking.template_args["batch"]["default"]
            )
        )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this plugin, you must have a ZulipChat bot defined; See here:
#  https://zulipchat.com/help/add-a-bot-or-integration
#
# At the time of writing this plugin the instructions were:
#    1. From your desktop, click on the gear icon in the upper right corner.
#    2. Select Settings.
#    3. On the left, click Your bots.
#    4. Click Add a new bot.
#    5. Fill out the fields, and click Create bot.

# If you know your organization {ID} (as it's part of the zulipchat.com url
# after you signup, then you can also access your bot information by visting:
#   https://ID.zulipchat.com/#settings/your-bots

# For example, I create an organization called apprise.  Thus my URL would be
#   https://apprise.zulipchat.com/#settings/your-bots

#  When you're done and have a bot, it's important to remember the username
#  you provided the bot and the API key generated.
#
#  If your {user} was   : goober-bot@apprise.zulipchat.com
#  and your {apikey} was: lqn6mpwpam6VZzbCW0o7olmk3hwbQSK
#
#  Then the following URLs would be accepted by Apprise:
#   - zulip://goober-bot@apprise.zulipchat.com/lqn6mpwpam6VZzbCW0o7olmk3hwbQSK
#   - zulip://goober-bot@apprise/lqn6mpwpam6VZzbCW0o7olmk3hwbQSK
#   - zulip://goober@apprise/lqn6mpwpam6VZzbCW0o7olmk3hwbQSK
#   - zulip://goober@apprise.zulipchat.com/lqn6mpwpam6VZzbCW0o7olmk3hwbQSK

# The API reference used to build this plugin was documented here:
#  https://zulipchat.com/api/send-message
#
import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import is_email, parse_list, validate_regex
from .base import NotifyBase

# A Valid Bot Name
VALIDATE_BOTNAME = re.compile(r"(?P<name>[A-Z0-9_-]{1,32})", re.I)

# Organization required as part of the API request
VALIDATE_ORG = re.compile(
    r"(?P<org>[A-Z0-9_-]{1,32})(\.(?P<hostname>[^\s]+))?", re.I
)

# Extend HTTP Error Messages
ZULIP_HTTP_ERROR_MAP = {
    401: "Unauthorized - Invalid Token.",
}

# Used to break path apart into list of streams
TARGET_LIST_DELIM = re.compile(r"[ \t\r\n,#\\/]+")

# Used to detect a streams
IS_VALID_TARGET_RE = re.compile(r"#?(?P<stream>[A-Z0-9_]{1,32})", re.I)


class NotifyZulip(NotifyBase):
    """A wrapper for Zulip Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Zulip"

    # The services URL
    service_url = "https://zulipchat.com/"

    # The default secure protocol
    secure_protocol = "zulip"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_zulip"

    # Zulip uses the http protocol with JSON requests
    notify_url = "https://{org}.{hostname}/api/v1/messages"

    # The maximum allowable characters allowed in the title per message
    title_maxlen = 60

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 10000

    # Define object templates
    templates = (
        "{schema}://{botname}@{organization}/{token}",
        "{schema}://{botname}@{organization}/{token}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "botname": {
                "name": _("Bot Name"),
                "type": "string",
                "regex": (r"^[A-Z0-9_-]{1,32}$", "i"),
                "required": True,
            },
            "organization": {
                "name": _("Organization"),
                "type": "string",
                "required": True,
                "regex": (r"^[A-Z0-9_-]{1,32})$", "i"),
            },
            "token": {
                "name": _("Token"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[A-Z0-9]{32}$", "i"),
            },
            "target_user": {
                "name": _("Target User"),
                "type": "string",
                "map_to": "targets",
            },
            "target_stream": {
                "name": _("Target Stream"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "token": {
                "alias_of": "token",
            },
        },
    )

    # The default hostname to append to a defined organization
    # if one isn't defined in the apprise url
    default_hostname = "zulipchat.com"

    # The default stream to notify if no targets are specified
    default_notification_stream = "general"

    def __init__(self, botname, organization, token, targets=None, **kwargs):
        """Initialize Zulip Object."""
        super().__init__(**kwargs)

        # our default hostname
        self.hostname = self.default_hostname

        try:
            match = VALIDATE_BOTNAME.match(botname.strip())
            if not match:
                # let outer exception handle this
                raise TypeError

            # The botname
            botname = match.group("name")
            suffix = "-bot"
            # Eliminate suffix if found
            botname = (
                botname[: -len(suffix)]
                if botname.endswith(suffix)
                else botname
            )
            self.botname = botname

        except (TypeError, AttributeError) as err:
            msg = f"The Zulip botname specified ({botname}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg) from err

        try:
            match = VALIDATE_ORG.match(organization.strip())
            if not match:
                # let outer exception handle this
                raise TypeError

            # The organization
            self.organization = match.group("org")
            if match.group("hostname"):
                self.hostname = match.group("hostname")

        except (TypeError, AttributeError) as err:
            msg = (
                "The Zulip organization specified "
                f"({organization}) is invalid."
            )
            self.logger.warning(msg)
            raise TypeError(msg) from err

        self.token = validate_regex(
            token, *self.template_tokens["token"]["regex"]
        )
        if not self.token:
            msg = f"The Zulip token specified ({token}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.targets = parse_list(targets)
        if len(self.targets) == 0:
            # No streams identified, use default
            self.targets.append(self.default_notification_stream)

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Zulip Notification."""

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
        }

        # error tracking (used for function return)
        has_error = False

        # Prepare our notification URL
        url = self.notify_url.format(
            org=self.organization,
            hostname=self.hostname,
        )

        # prepare JSON Object
        payload = {
            "subject": title,
            "content": body,
        }

        # Determine Authentication
        auth = (
            f"{self.botname}-bot@{self.organization}.{self.hostname}",
            self.token,
        )

        # Create a copy of the target list
        targets = list(self.targets)
        while len(targets):
            target = targets.pop(0)
            result = is_email(target)
            if result:
                # Send a private message
                payload["type"] = "private"
            else:
                # Send a stream message
                payload["type"] = "stream"

            # Set our target
            payload["to"] = target if not result else result["full_email"]

            self.logger.debug(
                f"Zulip POST URL: {url} "
                f"(cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"Zulip Payload: {payload!s}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    url,
                    data=payload,
                    headers=headers,
                    auth=auth,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyZulip.http_response_code_lookup(
                        r.status_code, ZULIP_HTTP_ERROR_MAP
                    )

                    self.logger.warning(
                        "Failed to send Zulip notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(f"Sent Zulip notification to {target}.")

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Zulip "
                    f"notification to {target}."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.organization,
            self.hostname,
            self.token,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        # simplify our organization in our URL if we can
        organization = "{}{}".format(
            self.organization,
            (
                f".{self.hostname}"
                if self.hostname != self.default_hostname
                else ""
            ),
        )

        return "{schema}://{botname}@{org}/{token}/{targets}?{params}".format(
            schema=self.secure_protocol,
            botname=NotifyZulip.quote(self.botname, safe=""),
            org=NotifyZulip.quote(organization, safe=""),
            token=self.pprint(self.token, privacy, safe=""),
            targets="/".join(
                [NotifyZulip.quote(x, safe="") for x in self.targets]
            ),
            params=NotifyZulip.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.targets)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The botname
        results["botname"] = NotifyZulip.unquote(results["user"])

        # The organization is stored in the hostname
        results["organization"] = NotifyZulip.unquote(results["host"])

        # Store our targets
        results["targets"] = NotifyZulip.split_path(results["fullpath"])

        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            # Store our token if specified
            results["token"] = NotifyZulip.unquote(results["qsd"]["token"])

        elif results["targets"]:
            # First item is the token
            results["token"] = results["targets"].pop(0)

        else:
            # no token
            results["token"] = None

        # Support the 'to' variable so that we can support rooms this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += list(
                filter(
                    bool,
                    TARGET_LIST_DELIM.split(
                        NotifyZulip.unquote(results["qsd"]["to"])
                    ),
                )
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from ..common import NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_bool
from .base import NotifyBase

# Default our global support flag
NOTIFY_GROWL_SUPPORT_ENABLED = False

try:
    import gntp.notifier

    # We're good to go!
    NOTIFY_GROWL_SUPPORT_ENABLED = True

except ImportError:
    # No problem; we just simply can't support this plugin until
    # gntp is installed
    pass


# Priorities
class GrowlPriority:
    LOW = -2
    MODERATE = -1
    NORMAL = 0
    HIGH = 1
    EMERGENCY = 2


GROWL_PRIORITIES = {
    # Note: This also acts as a reverse lookup mapping
    GrowlPriority.LOW: "low",
    GrowlPriority.MODERATE: "moderate",
    GrowlPriority.NORMAL: "normal",
    GrowlPriority.HIGH: "high",
    GrowlPriority.EMERGENCY: "emergency",
}

GROWL_PRIORITY_MAP = {
    # Maps against string 'low'
    "l": GrowlPriority.LOW,
    # Maps against string 'moderate'
    "m": GrowlPriority.MODERATE,
    # Maps against string 'normal'
    "n": GrowlPriority.NORMAL,
    # Maps against string 'high'
    "h": GrowlPriority.HIGH,
    # Maps against string 'emergency'
    "e": GrowlPriority.EMERGENCY,
    # Entries to additionally support (so more like Growl's API)
    "-2": GrowlPriority.LOW,
    "-1": GrowlPriority.MODERATE,
    "0": GrowlPriority.NORMAL,
    "1": GrowlPriority.HIGH,
    "2": GrowlPriority.EMERGENCY,
}


class NotifyGrowl(NotifyBase):
    """A wrapper to Growl Notifications."""

    # Set our global enabled flag
    enabled = NOTIFY_GROWL_SUPPORT_ENABLED

    requirements = {
        # Define our required packaging in order to work
        "packages_required": "gntp"
    }

    # The default descriptive name associated with the Notification
    service_name = "Growl"

    # The services URL
    service_url = "http://growl.info/"

    # The default protocol
    protocol = "growl"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_growl"

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_72

    # Disable throttle rate for Growl requests since they are normally
    # local anyway
    request_rate_per_sec = 0

    # Limit results to just the first 10 line otherwise there is just to much
    # content to display
    body_max_line_count = 2

    # Default Growl Port
    default_port = 23053

    # The Growl notification type used
    growl_notification_type = "New Messages"

    # Define object templates
    templates = (
        "{schema}://{host}",
        "{schema}://{host}:{port}",
        "{schema}://{password}@{host}",
        "{schema}://{password}@{host}:{port}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "priority": {
                "name": _("Priority"),
                "type": "choice:int",
                "values": GROWL_PRIORITIES,
                "default": GrowlPriority.NORMAL,
            },
            "version": {
                "name": _("Version"),
                "type": "choice:int",
                "values": (1, 2),
                "default": 2,
            },
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": True,
                "map_to": "include_image",
            },
            "sticky": {
                "name": _("Sticky"),
                "type": "bool",
                "default": True,
                "map_to": "sticky",
            },
        },
    )

    def __init__(
        self,
        priority=None,
        version=2,
        include_image=True,
        sticky=False,
        **kwargs,
    ):
        """Initialize Growl Object."""
        super().__init__(**kwargs)

        if not self.port:
            self.port = self.default_port

        # The Priority of the message
        self.priority = (
            NotifyGrowl.template_args["priority"]["default"]
            if not priority
            else next(
                (
                    v
                    for k, v in GROWL_PRIORITY_MAP.items()
                    if str(priority).lower().startswith(k)
                ),
                NotifyGrowl.template_args["priority"]["default"],
            )
        )

        # Our Registered object
        self.growl = None

        # Sticky flag
        self.sticky = sticky

        # Store Version
        self.version = version

        # Track whether or not we want to send an image with our notification
        # or not.
        self.include_image = include_image

        return

    def register(self):
        """Registers with the Growl server."""
        payload = {
            "applicationName": self.app_id,
            "notifications": [
                self.growl_notification_type,
            ],
            "defaultNotifications": [
                self.growl_notification_type,
            ],
            "hostname": self.host,
            "port": self.port,
        }

        if self.password is not None:
            payload["password"] = self.password

        self.logger.debug(f"Growl Registration Payload: {payload!s}")
        self.growl = gntp.notifier.GrowlNotifier(**payload)

        try:
            self.growl.register()

        except gntp.errors.NetworkError:
            msg = (
                "A network error error occurred registering "
                f"with Growl at {self.host}."
            )
            self.logger.warning(msg)
            return False

        except gntp.errors.ParseError:
            msg = (
                "A parsing error error occurred registering "
                f"with Growl at {self.host}."
            )
            self.logger.warning(msg)
            return False

        except gntp.errors.AuthError:
            msg = (
                "An authentication error error occurred registering "
                f"with Growl at {self.host}."
            )
            self.logger.warning(msg)
            return False

        except gntp.errors.UnsupportedError:
            msg = (
                "An unsupported error occurred registering with "
                f"Growl at {self.host}."
            )
            self.logger.warning(msg)
            return False

        self.logger.debug("Growl server registration completed successfully.")

        # Return our state
        return True

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Growl Notification."""
        # Register ourselves with the server if we haven't done so already
        if not self.growl and not self.register():
            # We failed to register
            return False

        icon = None
        if self.version >= 2:
            # URL Based
            icon = (
                None if not self.include_image else self.image_url(notify_type)
            )

        else:
            # Raw
            icon = (
                None if not self.include_image else self.image_raw(notify_type)
            )

        payload = {
            "noteType": self.growl_notification_type,
            "title": title,
            "description": body,
            "icon": icon is not None,
            "sticky": self.sticky,
            "priority": self.priority,
        }
        self.logger.debug(f"Growl Payload: {payload!s}")

        # Update icon of payload to be raw data; this is intentionally done
        # here after we spit the debug message above (so we don't try to
        # print the binary contents of an image
        payload["icon"] = icon

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            # Perform notification
            response = self.growl.notify(**payload)
            if not isinstance(response, bool):
                self.logger.warning(
                    "Growl notification failed to send with response:"
                    f" {response!s}",
                )

            else:
                self.logger.info("Sent Growl notification.")

        except gntp.errors.BaseError as e:
            # Since Growl servers listen for UDP broadcasts, it's possible
            # that you will never get to this part of the code since there is
            # no acknowledgement as to whether it accepted what was sent to it
            # or not.

            # However, if the host/server is unavailable, you will get to this
            # point of the code.
            self.logger.warning(
                "A Connection error occurred sending Growl "
                f"notification to {self.host}."
            )
            self.logger.debug(f"Growl Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port if self.port else self.default_port,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "image": "yes" if self.include_image else "no",
            "sticky": "yes" if self.sticky else "no",
            "priority": (
                GROWL_PRIORITIES[self.template_args["priority"]["default"]]
                if self.priority not in GROWL_PRIORITIES
                else GROWL_PRIORITIES[self.priority]
            ),
            "version": self.version,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        auth = ""
        if self.user:
            # The growl password is stored in the user field
            auth = "{password}@".format(
                password=self.pprint(
                    self.user, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )

        return "{schema}://{auth}{hostname}{port}/?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port=(
                ""
                if self.port is None or self.port == self.default_port
                else f":{self.port}"
            ),
            params=NotifyGrowl.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        version = None
        if "version" in results["qsd"] and len(results["qsd"]["version"]):
            # Allow the user to specify the version of the protocol to use.
            try:
                version = int(
                    NotifyGrowl.unquote(results["qsd"]["version"])
                    .strip()
                    .split(".")[0]
                )

            except (AttributeError, IndexError, TypeError, ValueError):
                NotifyGrowl.logger.warning(
                    'An invalid Growl version of "{}" was specified and will '
                    "be ignored.".format(results["qsd"]["version"])
                )
                pass

        # Set our priority
        if "priority" in results["qsd"] and len(results["qsd"]["priority"]):
            results["priority"] = NotifyGrowl.unquote(
                results["qsd"]["priority"]
            )

        # Because of the URL formatting, the password is actually where the
        # username field is. For this reason, we just preform this small hack
        # to make it (the URL) conform correctly. The following strips out the
        # existing password entry (if exists) so that it can be swapped with
        # the new one we specify.
        if results.get("password", None) is None:
            results["password"] = results.get("user", None)

        # Include images with our message
        results["include_image"] = parse_bool(
            results["qsd"].get(
                "image", NotifyGrowl.template_args["image"]["default"]
            )
        )

        # Include images with our message
        results["sticky"] = parse_bool(
            results["qsd"].get(
                "sticky", NotifyGrowl.template_args["sticky"]["default"]
            )
        )

        # Set our version
        if version:
            results["version"] = version

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import inspect

from .. import common
from ..logger import logger
from ..manager_plugins import NotificationManager
from ..plugins.base import NotifyBase
from ..utils.logic import dict_full_update
from ..utils.parse import URL_DETAILS_RE, parse_url, url_assembly

# Grant access to our Notification Manager Singleton
N_MGR = NotificationManager()


class CustomNotifyPlugin(NotifyBase):
    """Apprise Custom Plugin Hook.

    This gets initialized based on @notify decorator definitions
    """

    # Our Custom notification
    service_url = "https://github.com/caronc/apprise/wiki/Custom_Notification"

    # Over-ride our category since this inheritance of the NotifyBase class
    # should be treated differently.
    category = "custom"

    # Support Attachments
    attachment_support = True

    # Allow persistent storage support
    storage_mode = common.PersistentStoreMode.AUTO

    # Define object templates
    templates = ("{schema}://",)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns arguments retrieved."""
        return parse_url(url, verify_host=False, simple=True)

    def url(self, privacy=False, *args, **kwargs):
        """General URL assembly."""
        return f"{self.secure_protocol}://"

    @staticmethod
    def instantiate_plugin(url, send_func, name=None):
        """The function used to add a new notification plugin based on the
        schema parsed from the provided URL into our supported matrix
        structure."""

        if not isinstance(url, str):
            msg = (
                f"An invalid custom notify url/schema ({url}) provided in "
                f"function {send_func.__name__}."
            )
            logger.warning(msg)
            return None

        # Validate that our schema is okay
        re_match = URL_DETAILS_RE.match(url)
        if not re_match:
            msg = (
                f"An invalid custom notify url/schema ({url}) provided in "
                f"function {send_func.__name__}."
            )
            logger.warning(msg)
            return None

        # Acquire our schema
        schema = re_match.group("schema").lower()

        if not re_match.group("base"):
            url = f"{schema}://"

        # Keep a default set of arguments to apply to all called references
        base_args = parse_url(
            url, default_schema=schema, verify_host=False, simple=True
        )

        if schema in N_MGR:
            # we're already handling this object
            msg = (
                f"The schema ({url}) is already defined and could not be "
                f"loaded from custom notify function {send_func.__name__}."
            )
            logger.warning(msg)
            return None

        # We define our own custom wrapper class so that we can initialize
        # some key default configuration values allowing calls to our
        # `Apprise.details()` to correctly differentiate one custom plugin
        # that was loaded from another
        class CustomNotifyPluginWrapper(CustomNotifyPlugin):

            # Our Service Name
            service_name = (
                name
                if isinstance(name, str) and name
                else f"Custom - {schema}"
            )

            # Store our matched schema
            secure_protocol = schema

            requirements = {
                # Define our required packaging in order to work
                "details": f"Source: {inspect.getfile(send_func)}"
            }

            # Assign our send() function
            __send = staticmethod(send_func)

            # Update our default arguments
            _base_args = base_args

            def __init__(self, **kwargs):
                """Our initialization."""
                #  init parent
                super().__init__(**kwargs)

                self._default_args = {}

                # Some variables do not need to be set
                kwargs.pop("secure", None)

                # Apply our updates based on what was parsed
                dict_full_update(self._default_args, self._base_args)
                dict_full_update(self._default_args, kwargs)

                # Update our arguments (applying them to what we originally)
                # initialized as
                self._default_args["url"] = url_assembly(**self._default_args)

            def send(
                self,
                body,
                title="",
                notify_type=common.NotifyType.INFO,
                *args,
                **kwargs,
            ):
                """Our send() call which triggers our hook."""

                response = False
                try:
                    # Enforce a boolean response
                    result = self.__send(
                        body,
                        title,
                        notify_type.value,
                        *args,
                        meta=self._default_args,
                        **kwargs,
                    )

                    # None and True are both considered successful
                    # False however is passed along further upstream
                    response = True if result is None else bool(result)

                except Exception as e:
                    # Unhandled Exception
                    self.logger.warning(
                        "An exception occured sending a %s notification.",
                        N_MGR[self.secure_protocol].service_name,
                    )
                    self.logger.debug(
                        "%s Exception: %s", N_MGR[self.secure_protocol], str(e)
                    )
                    return False

                if response:
                    self.logger.info(
                        "Sent %s notification.",
                        N_MGR[self.secure_protocol].service_name,
                    )
                else:
                    self.logger.warning(
                        "Failed to send %s notification.",
                        N_MGR[self.secure_protocol].service_name,
                    )
                return response

        # Store our plugin into our core map file
        return N_MGR.add(
            plugin=CustomNotifyPluginWrapper,
            schemas=schema,
            send_func=send_func,
            url=url,
        )



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Feishu
#   1. Visit https://open.feishu.cn

# Custom Bot Setup
#    https://open.feishu.cn/document/client-docs/bot-v3/add-custom-bot
#

from json import dumps

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase


class NotifyFeishu(NotifyBase):
    """A wrapper for Feishu Notifications."""

    # The default descriptive name associated with the Notification
    service_name = _("Feishu")

    # The services URL
    service_url = "https://open.feishu.cn/"

    # The default secure protocol
    secure_protocol = "feishu"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_feishu"

    # Notification URL
    notify_url = "https://open.feishu.cn/open-apis/bot/v2/hook/{token}/"

    # Define object templates
    templates = ("{schema}://{token}",)

    # The title is not used
    title_maxlen = 0

    # Limit is documented to be 20K message sizes.  This number safely
    # allows padding around that size.
    body_maxlen = 19985

    # Define our tokens; these are the minimum tokens required required to
    # be passed into this function (as arguments). The syntax appends any
    # previously defined in the base package and builds onto them
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("Token"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[A-Z0-9_-]+$", "i"),
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "token": {
                "alias_of": "token",
            },
        },
    )

    def __init__(self, token, **kwargs):
        """Initialize Feishu Object."""
        super().__init__(**kwargs)

        self.token = validate_regex(
            token, *self.template_tokens["token"]["regex"]
        )
        if not self.token:
            msg = f"The Feishu token specified ({token}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Send our notification."""

        # prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # Our Message
        payload = {
            "msg_type": "text",
            "content": {
                "text": body,
            },
        }

        self.logger.debug(
            "Feishu GET URL:"
            f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Feishu Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                self.notify_url.format(token=self.token),
                data=dumps(payload).encode("utf-8"),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            #
            # Sample Responses
            #
            # Valid:
            # {
            #   "code": 0,
            #   "data": {},
            #   "msg": "success"
            # }

            # Invalid (non 200 response):
            # {
            #   "code": 9499,
            #   "msg": "Bad Request",
            #   "data": {}
            # }
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyFeishu.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Feishu notification: {}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Feishu notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Feishu notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.token)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Prepare our parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return "{schema}://{token}/?{params}".format(
            schema=self.secure_protocol,
            token=self.pprint(self.token, privacy, safe=""),
            params=NotifyFeishu.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        # parse_url already handles getting the `user` and `password` fields
        # populated.
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Allow over-ride
        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            results["token"] = NotifyFeishu.unquote(results["qsd"]["token"])

        else:
            results["token"] = NotifyFeishu.unquote(results["host"])

        return results
